class block_set:
    """A decorator that validates that required context keys are present.

    Use this decorator to wrap a function that renders a block set for Slack.
    This will check that the context provided to the renderer contains all the
    required context variables.

    Args:
        required_context (list): List of keys to look up in context.
    """

    def __init__(self, required_context=[]):
        self.required_context = required_context

    def __call__(self, f):
        def wrapped_f(context):
            for prop in self.required_context:
                if context.get(prop) is None:
                    raise ValueError(f"context missing: {prop}, in {f.__name__}")
            return f(context)

        return wrapped_f


class processor:
    """
    Decorator. Checks for required context for a processor.
    """

    def __init__(self, required_context=[], *args, **kwargs):
        self.required_context = required_context

    def __call__(self, f):
        def wrapped_f(payload, context, *args, **kwargs):
            for prop in self.required_context:
                if context.get(prop) is None:
                    raise ValueError(f"context missing: {prop}, in {f.__name__}")
            return f(payload, context, *args, **kwargs)

        return wrapped_f



import uuid


class BlockBuilder:
    @staticmethod
    def simple_context_block(value, text_type="plain_text", block_id=None):
        """Simple single value context block - metadata style

        Parameters
        -----------
        value - a text value
        text_type - if using markdown set to mrkdwn
        block_id: str
            block i.d number. If not entered will generate random str
        """

        if not block_id:
            block_id = str(uuid.uuid4())
        return {
            "type": "context",
            "elements": [{"type": text_type, "text": value}],
            "block_id": block_id,
        }

    def many_context_block(value, text_type="plain_text", block_id=None):
        """Simple multi context block - metadata style

        Parameters
        -----------
        value - [a text value]
        text_type - if using markdown set to mrkdwn this is applied to all instances
        block_id: str
            block i.d number. If not entered will generate random str
        """

        if not isinstance(value, list):
            raise Exception("Please use an array of text values if you do not want multiple use the simple_context_block")
        if not block_id:
            block_id = str(uuid.uuid4())
        elements = [{"type": text_type, "text": val} for val in value]
        return {
            "type": "context",
            "elements": elements,
            "block_id": block_id,
        }

    def custom_context_block(values=[], block_id=None):
        """Customizable multi context block - metadata style

        Parameters
        -----------
        values=[] - an array of text_type (see text_block )
        block_id: str
            block i.d number. If not entered will generate random str
        """

        if not isinstance(values, list):
            raise Exception("Please use an array of text values if you do not want multiple use the simple_context_block")
        if not block_id:
            block_id = str(uuid.uuid4())
        elements = values
        return {
            "type": "context",
            "elements": elements,
            "block_id": block_id,
        }

    @staticmethod
    def text_block(value, text_type="plain_text"):
        return {"type": text_type, "text": value}

    @staticmethod
    def input_block(
        label,
        initial_value=None,
        placeholder=False,
        multiline=False,
        placeholder_type="plain_text",
        action_id="plain_input",
        block_id=None,
        label_type="plain_text",
        min_length=False,
        max_length=False,
        optional=True,
    ):
        """
        Creates a basic text input block

        Parameters:
        label - String of what the label text should be
        initial_value - Set an initial value for input using string passed in, otherwise None
        placeholder - Sets placeholder text for input using string passed in, defaults to no placeholder
        multiline - If True will set text input (in slack) to a Multiline text area
        placeholder_type - Only used if placeholder is passed in, sets placeholder type to plain_text
        action_id - Sets id of input to value passed in, otherwise defaults to 'plain_input'
        block_id - Set id of block to value passed in, if none is passed in will create one with uuid
        label_type - Sets the label type to plain_text. SHOULD NOT BE CHANGED.
        min_length - Set a minimum length for the input
        max_length - Set a maximum length for the input
        optional - indicated whether the input can be left blank

        If a placeholder, min_length, max_length is
        passed in it will be used otherwise False
        will not add a placeholder

        For a basic input only a string for the label has to be passed in.
        """

        if not block_id:
            block_id = str(uuid.uuid4())
        obj = {
            "type": "input",
            "block_id": block_id,
            "label": {"type": label_type, "text": label},
            "optional": optional,
            "element": {
                "type": "plain_text_input",
                "action_id": action_id,
                "multiline": multiline,
            },
        }
        if placeholder:
            # placeholder is a text_block
            obj["element"]["placeholder"] = BlockBuilder.text_block(placeholder, placeholder_type)

        if max_length:
            obj["element"]["max_length"] = max_length

        if min_length:
            obj["element"]["min_length"] = min_length

        if initial_value:
            initial_value = str(initial_value)
            obj["element"]["initial_value"] = initial_value

        return obj

    @staticmethod
    def simple_section_block(value, text_type="plain_text", block_id=None):
        if not block_id:
            block_id = str(uuid.uuid4())
        return {
            "type": "section",
            "text": {"type": text_type, "text": value},
            "block_id": block_id,
        }

    @staticmethod
    def simple_section_multiple_block(text_blocks, block_id=None):
        """sections can have multiple fields they are a collection of text_block

        Parameters
        -----------
        text_blocks: object
            contains plain text or markup
        block_id: str
            block i.d number. If not entered will generate random str
        """

        if not block_id:
            block_id = str(uuid.uuid4())
        return {"type": "section", "fields": text_blocks, "block_id": block_id}

    @staticmethod
    def option(text, value):
        """accepts a string and returns an object with 2 properties, type and text.

        The type property is set to plain_text. The text property is given the value of the
        string entered.

        Parameters
        ----------
        text: str
            plain text
        value: any
        """

        return {
            "text": {"type": "plain_text", "text": text},
            "value": value,
        }

    @staticmethod
    def divider_block():
        """returns an object that has one property, type:divider"""
        return {"type": "divider"}

    @staticmethod
    def header_block(text, block_id=None):
        """
        Creates a header block to be used inside other blocks

        Parameters:
        text - Set text as any String passed in.
        block_id - Optional Id can be passed in, otherwise creates one with uuid
        """
        if not block_id:
            block_id = str(uuid.uuid4())
        return {
            "type": "header",
            "text": {"type": "plain_text", "text": text},
            "block_id": block_id,
        }

    @staticmethod
    def external_select_block(
        label,
        action_id,
        initial_option=None,
        block_id=None,
        min_query_length=0,
        placeholder="Select",
    ):
        """returns a section block with markdown text and an external_select dropdown menu

        External menu's load their options from an external data source, allowing for a dynamic list of options.

        Parameters
        ----------
        label: mrkdwn
            markdown text for the section block.
        action_id: str
            An identifier for the action triggered when a menu option is selected.
        initial_option: obj
            The option selected when the menu initially loads.
        block_id: str
            Unique identifier for the block
        min_query_length: int
            Tell's Slack the fewest number of typed characters required before dispatch.
        placeholder: obj
            Defines the placeholder text shown on the menu. Maximum length for the text in this field is 150 characters.
        """

        block = {
            "type": "section",
            "text": {"type": "mrkdwn", "text": f"{label}"},
            "accessory": {
                "type": "external_select",
                "action_id": action_id,
                "placeholder": {"type": "plain_text", "text": placeholder},
                "min_query_length": min_query_length,
            },
        }
        if initial_option:
            block["accessory"]["initial_option"] = initial_option
        if block_id:
            block["block_id"] = block_id
        return block

    @staticmethod
    def static_select_block(
        label,
        options,
        action_id=None,
        initial_option=None,
        placeholder="Select",
        block_id=None,
    ):
        """ "
        Returns a select block where the select values are known and won't change.

        Parameters:
        label - A string for the select input label
        options - Options are an array of options (see above)
        action_id - Will be included if entered otherwise None
        initial_option - An option block has to be entered, this shows the entered option
        when first being displayed
        placeholder - Changed placeholder text, otherwise defautls to Select
        block_id - Sets block_id, otherwise will use uuid to set.

        Function returns an entire section with the select input.
        Only parameters that need to be entered are label and array of options.
        """

        if not block_id:
            block_id = str(uuid.uuid4())
        block = {
            "type": "section",
            "text": {"type": "mrkdwn", "text": f"{label}"},
            "block_id": block_id,
            "accessory": {
                "type": "static_select",
                "placeholder": {"type": "plain_text", "text": placeholder},
                "options": options,
            },
        }
        if initial_option:
            block["accessory"]["initial_option"] = initial_option
        if action_id:
            block["accessory"]["action_id"] = action_id
        return block

    @staticmethod
    def multi_static_select_block(
        label,
        options,
        action_id=None,
        initial_options=None,
        placeholder="Select",
        block_id=None,
    ):
        """
        Creates a selection input for selecting multiple options

        Parameters:
        label - String of what the label text should be
        options  - must be an array of options (see above)
        action_id - set id to the value passed in, otherwise None
        initial_options - must be an array of options (see above) to show when initially loading
        placeholder - String for the placeholder text in the input, otherwise 'Select'
        block_id - set id of block to value passed in, otherwise created with uuid

        Should be used when the value of the selection are known
        and do not need to be gotten from and external source
        """

        if not block_id:
            block_id = str(uuid.uuid4())
        block = {
            "type": "section",
            "text": {"type": "mrkdwn", "text": f"{label}"},
            "block_id": block_id,
            "accessory": {
                "type": "multi_static_select",
                "placeholder": {"type": "plain_text", "text": placeholder},
                "options": options,
            },
        }
        if initial_options:
            block["accessory"]["initial_options"] = initial_options
        if action_id:
            block["accessory"]["action_id"] = action_id
        return block

    @staticmethod
    def multi_external_select_block(
        label,
        action_id,
        initial_options=None,
        placeholder="Select",
        block_id=None,
        min_query_length=0,
    ):
        """A section block with markup text and a multi_external_select menu.

        Works the same as the external_select, but allows a user to select multiple items.

        If the section block has no id it will generate a random one. Returns the section block
        with the initial option if one was selected, and without the initial option if it was not
        selected.
        """

        if not block_id:
            block_id = str(uuid.uuid4())
        block = {
            "type": "section",
            "text": {"type": "mrkdwn", "text": f"{label}"},
            "block_id": block_id,
            "accessory": {
                "type": "multi_external_select",
                "placeholder": {"type": "plain_text", "text": placeholder},
                "action_id": action_id,
                "min_query_length": min_query_length,
            },
        }
        if initial_options:
            block["accessory"]["initial_options"] = initial_options

        return block

    @staticmethod
    def datepicker_block(
        initial_date=None,
        action_id=None,
        block_id=None,
        label="Select Date",
        placeholder="Select a date",
    ):
        """
        Function returns a datepicker block object

        Parameters:
        initial_date - Sets the initial value to the date string passed in, must be 'YEAR-MONTH-DAY' format (ex '2020-04-21')
        action_id - Sets id to value passed in, otherwise None
        block_id - Sets id to value passed in, otherwise sets id with uuid
        label - Sets the label to the string passed in, otherwise default is 'Select Date'
        placeholder - Sets the placeholder text of the input to string passed in, otherwise default is 'Select a date'

        """
        if not block_id:
            block_id = str(uuid.uuid4())
        block = {
            "type": "section",
            "text": {"type": "mrkdwn", "text": f"{label}"},
            "block_id": block_id,
            "accessory": {
                "type": "datepicker",
                "placeholder": {"type": "plain_text", "text": f"{placeholder}"},
            },
        }
        if initial_date:
            block["accessory"]["initial_date"] = initial_date
        if action_id:
            block["accessory"]["action_id"] = action_id
        return block

    @staticmethod
    def section_with_button_block(
        button_label,
        button_value,
        section_text,
        text_type="mrkdwn",
        block_id=None,
        url=None,
        style=None,
        confirm=False,
        action_id=None,
    ):
        """
        Function returns a section object with a button.

        Parameters:
        button_label - Sets the text of the button, needs to be included
        button_value - Sets the value to be passed along in the payload, needs to be included
        section_text - Sets the text of the section body, needs to be included
        text_type - Sets the type for the section block, generally doesn't need to be changed
        block_id - Sets the block_id to string passed in, otherwise is set with uuid
        url - If included the button will forward the user to the url, otherwise None
        style - If included changes the style of the button, options are default, primary, or danger.
        confirm - If True will have a popup confirmation window appear when button is pressed.
        action_id - Sets the action_id to string passed in, otherwise is set with uuid.

        Used for linking to another site/page or confirming information.
        """
        if not block_id:
            block_id = str(uuid.uuid4())
        block = {
            "type": "section",
            "text": {"type": text_type, "text": section_text},
            "block_id": block_id,
            "accessory": {
                "type": "button",
                "text": {"type": "plain_text", "text": button_label},
                "value": button_value,
                # "confirm": confirm,
                "action_id": action_id if action_id else str(uuid.uuid4()),
            },
        }
        if style:
            block["accessory"]["style"] = style
        if url:
            block["accessory"]["url"] = url
        return block

    @staticmethod
    def simple_button_block(label, value, url=None, style=None, confirm=False, action_id=None):
        """
        Function returns an object for a button block.

        label - Sets text of the button
        value - Sets value of the button to be passed in the payload
        url - If included button will forward user to the url, otherwise None
        style - If included will change the appearance of the button.
                Choices are default, primary, or danger
        confirm - If included will add a popup window confirming button press
        action_id - Sets the action_id to string passed in, otherwise set with uuid, must be unique

        This block can only be used in action_block function and action_id must be unique
        """

        block = {
            "type": "button",
            "text": {"type": "plain_text", "text": label},
            "value": value,
            # "confirm": confirm,
            "action_id": action_id if action_id else str(uuid.uuid4()),
        }
        if style:
            block["style"] = style
        if url:
            block["url"] = url
        return block

    @staticmethod
    def actions_block(blocks=[], block_id=None):
        """
        Array of interactive element objects - buttons, select menus, overflow menus, or date pickers.
        max of 5

        parameters
        ----------
        blocks: array
            Element objects
        block_id: str
            A unique identifier for the block
        """

        if not len(blocks):
            return
        if len(blocks) > 4:
            return
        if not block_id:
            block_id = str(uuid.uuid4())
        return {"type": "actions", "block_id": block_id, "elements": [*blocks]}

    @staticmethod
    def checkbox_block(label, options, action_id=None, initial_options=None, block_id=None):
        """
        Function returns a section with checkbox inputs.

        Parameters:
        label - String to set the text of the section, must be included
        options - array of options (see above), must be included
        action_id - sets action_id to value passed in, otherwise sets id from uuid
        initial_options - Array of options (see above), if included will pre-check those options
        block_id - sets block_is as value entered, otherwise sets if from uuid

        """
        if not action_id:
            action_id = str(uuid.uuid4())
        if not block_id:
            block_id = str(uuid.uuid4())
        block = {
            "type": "section",
            "block_id": block_id,
            "text": {"type": "mrkdwn", "text": f"{label}"},
            "accessory": {
                "type": "checkboxes",
                "action_id": action_id,
                "options": options,
            },
        }
        if initial_options:
            block["accessory"]["initial_options"] = initial_options
        return block

    @staticmethod
    def section_with_accessory_block(
        section_text,
        accessory,
        text_type="mrkdwn",
        block_id=None,
    ):
        """ " Builds a section with an accessory (image/button)

        parameters
        ----------
        section_text: str
            text for the section block
        accessory: obj
            element object
        text_type: str
            markdown
        block_id: str
            unique identifier
        """
        if not block_id:
            block_id = str(uuid.uuid4())
        block = {
            "type": "section",
            "text": {"type": text_type, "text": section_text},
            "block_id": block_id,
            "accessory": accessory,
        }

        return block

    @staticmethod
    def simple_image_block(url, alt_text):
        """An image block object with the image's url and alt text.

        parameters
        ----------
        url: str
            The URL of the image to be displayed. Maximum length for this field is 3000 characters.
        alt_text: str
            A plain-text summary of the image. This should not contain any markup. Maximum length for this field is 2000 characters.
        """
        return {
            "type": "image",
            "image_url": url,
            "alt_text": alt_text,
        }

    @staticmethod
    def multi_channels_select_block(
        section_text,
        placeholder_text="Select Channels",
        section_text_type="mrkdwn",
        block_id=None,
        action_id=None,
        initial_channels=[],
        max_selected_items=None,
    ):
        if not block_id:
            block_id = str(uuid.uuid4())
        accessory_block = {
            "type": "multi_channels_select",
            "placeholder": BlockBuilder.text_block(placeholder_text),
        }
        if action_id:
            accessory_block["action_id"] = action_id
        if initial_channels and len(initial_channels):
            accessory_block["initial_channels"] = initial_channels
        if max_selected_items not in ["", None]:
            accessory_block["max_selected_items"] = max_selected_items

        block = BlockBuilder.section_with_accessory_block(
            section_text,
            accessory_block,
            text_type=section_text_type,
            block_id=block_id,
        )
        return block

    @staticmethod
    def multi_conversations_select_block(
        section_text,
        placeholder_text="Select Channels",
        section_text_type="mrkdwn",
        block_id=None,
        action_id=None,
        initial_conversations=[],
        max_selected_items=None,
        default_to_current_conversation=False,
        filter_opts=None,  # see https://api.slack.com/reference/block-kit/composition-objects#filter_conversations
    ):
        if not block_id:
            block_id = str(uuid.uuid4())
        accessory_block = {
            "type": "multi_conversations_select",
            "placeholder": BlockBuilder.text_block(placeholder_text),
        }
        if action_id:
            accessory_block["action_id"] = action_id
        if initial_conversations and len(initial_conversations):
            accessory_block["initial_conversations"] = initial_conversations
        if max_selected_items not in ["", None]:
            accessory_block["max_selected_items"] = max_selected_items
        if filter_opts:
            accessory_block["filter"] = filter_opts

        block = BlockBuilder.section_with_accessory_block(
            section_text,
            accessory_block,
            text_type=section_text_type,
            block_id=block_id,
        )
        return block



import hashlib
import hmac
import json
import logging
import math
from datetime import datetime
from urllib.parse import urlencode

import requests
from requests.exceptions import HTTPError

from . import constants as slack_consts
from .block_builder import BlockBuilder
from .exceptions import SlackAppException
from .utils import SlackUtils

logger = logging.getLogger(__name__)


class SlackApp:
    SLACK_API_ROOT = "https://slack.com/api/"
    SLACK_OAUTH_AUTHORIZE_ROOT = "https://slack.com/oauth/v2/authorize"
    # https://api.slack.com/methods/oauth.v2.access
    OAUTH_V2_ACCESS = "oauth.v2.access"

    # https://api.slack.com/methods/conversations.open
    CONVERSATIONS_OPEN = "conversations.open"

    # https://api.slack.com/methods/chat.postMessage
    POST_MESSAGE = "chat.postMessage"

    POST_EPHEMERAL = "chat.postEphemeral"

    # https://api.slack.com/methods/views.open

    VIEWS_OPEN = "views.open"

    VIEWS_UPDATE = "views.update"

    VIEWS_PUSH = "views.push"

    VIEWS_PUBLISH = "views.publish"

    #  https://api.slack.com/methods/chat.getPermalink
    CHAT_GET_PERMALINK = "chat.getPermalink"

    # https://api.slack.com/methods/chat.update
    CHAT_UPDATE = "chat.update"

    APP_VERSION = "v0"

    TOKEN_TYPE_BOT = "bot"

    TOKEN_TYPE_USER = "user"

    JSON_HEADERS = {
        "Content-Type": "application/json; charset=utf-8",
        "Accept": "application/json",
    }
    URL_ENCODED_HEADERS = {
        "Accept": "application/json",
        "Content-Type": "application/x-www-form-urlencoded; charset=utf-8",
    }
    ROUTES = {
        slack_consts.BLOCK_ACTIONS: {},
        slack_consts.BLOCK_SUGGESTION: {},
        slack_consts.VIEW_SUBMISSION: {},
        slack_consts.VIEW_CLOSED: {},
    }

    _ROUTE_TYPES = [
        slack_consts.BLOCK_ACTIONS,
        slack_consts.BLOCK_SUGGESTION,
        slack_consts.VIEW_CLOSED,
        slack_consts.VIEW_SUBMISSION,
    ]
    BLOCK_SETS = dict()

    def __init__(
        self,
        client_id,
        client_secret,
        signing_secret,
        redirect_url,
        bot_scopes=[],
        user_scopes=[],
        error_webhook=None,
    ):
        self.CLIENT_ID = client_id
        self.CLIENT_SECRET = client_secret
        self.SIGNING_SECRET = signing_secret
        self.BOT_SCOPES = bot_scopes
        self.USER_SCOPES = user_scopes
        self.REDIRECT_URL = redirect_url
        self.ERROR_WEBHOOK = error_webhook

    @classmethod
    def configure(
        cls,
        client_id,
        client_secret,
        signing_secret,
        redirect_url,
        bot_scopes=[],
        user_scopes=[],
        error_webhook=None,
    ):
        return cls(
            client_id,
            client_secret,
            signing_secret,
            redirect_url,
            bot_scopes,
            user_scopes,
            error_webhook,
        )

    def _handle_response(self, response, fn_name=None, blocks=[]):
        if not hasattr(response, "status_code"):
            raise ValueError

        else:
            status_code = response.status_code
            try:
                res_data = response.json()
            except json.decoder.JSONDecodeError:
                # one slack request (see generic requests) does not return json
                return response.text

            if not res_data.get("ok"):
                error_code = response.status_code
                error_param = res_data.get("error")
                if res_data.get("response_metadata", None):
                    error_message = res_data.get("response_metadata", {}).get("messages")
                elif res_data.get("error", None):
                    error_message = res_data.get("error")
                else:
                    error_message = response.text

                kwargs = {
                    "status_code": status_code,
                    "error_code": error_code,
                    "error_param": error_param,
                    "error_message": error_message,
                }
                try:
                    SlackAppException(HTTPError(kwargs), blocks=blocks)
                except Exception as e:
                    if self.ERROR_WEBHOOK:
                        try:
                            self.generic_request(
                                self.ERROR_WEBHOOK,
                                {"text": f"An error occured {e}"},
                            )
                        except Exception as fail_safe_error:
                            logger.info(fail_safe_error)
                            pass
                    raise e

            return res_data

    def _check_request_time_stamp(self, request):
        time_stamp = request.headers.get("X-Slack-Request-Timestamp", None)
        if not time_stamp:
            raise ValueError
        is_expired = int(time_stamp) <= math.floor(datetime.timestamp(datetime.now() - datetime.timedelta(minutes=5)))
        if is_expired:
            raise ValueError

        return time_stamp

    def authenticate_incoming_request(self, request):

        time_stamp = self._check_time_stamp(request)
        slack_signature = request.headers.get("X-Slack-Signature", None)
        if not slack_signature:
            raise ValueError
        data = request.body.decode("utf-8")
        sig_basedstring = (f"{self.APP_VERSION}:{time_stamp}:{data}").encode("utf-8")
        my_sig = (
            self.APP_VERSION
            + "="
            + hmac.new(
                self.SIGNING_SECRET.encode("utf-8"),
                sig_basedstring,
                hashlib.sha256,
            ).hexdigest()
        )
        if hmac.compare_digest(my_sig, slack_signature):
            return True
        return False

        raise ValueError

    def retrieve_data_with_code(self, code):
        url = self.SLACK_API_ROOT + self.OAUTH_V2_ACCESS
        data = {
            "code": code,
            "redirect_uri": self.REDIRECT_URL,
            "client_id": self.CLIENT_ID,
            "client_secret": self.CLIENT_SECRET,
        }

        return self._handle_response(
            requests.post(
                url,
                data=data,
                headers=self.URL_ENCODED_HEADERS,
            )
        )

    def revoke_access_token(self, token):
        query = urlencode({"token": token})
        url = self.SLACK_API_ROOT + "auth.revoke?" + query

        return requests.post(
            url,
            headers=self.URL_ENCODED_HEADERS,
        )

    def block_finder(self, block_id, blocks=[]):
        item = list(
            filter(
                lambda x: x[1]["block_id"] == block_id,
                enumerate(blocks),
            )
        )
        if len(item):
            return item[0]
        return item

    def auth_headers(self, access_token):
        return {"Authorization": "Bearer " + access_token, **self.JSON_HEADERS}

    def _block_action_consumer(self, payload):
        action_query_string = payload["actions"][0]["action_id"]
        processed_string = self.utils.process_action_id(action_query_string)
        action_id = processed_string.get("true_id")
        action_params = processed_string.get("params")
        # special override for block_actions
        if action_params.get("__block_action", None):
            action_id = action_params.get("__block_action")
        switcher = self._action_routes
        return switcher.get(action_id, self.utils.NO_OP)(payload, action_params)

    def _block_suggestion_consumer(self, payload):
        action_query_string = payload["action_id"]
        processed_string = self.utils.process_action_id(action_query_string)
        action_id = processed_string.get("true_id")
        action_params = processed_string.get("params")
        switcher = self._suggestion_routes
        return switcher.get(action_id, self.utils.NO_OP)(payload, action_params)

    def _view_submission_consumer(self, payload):
        callback_id = payload["view"]["callback_id"]
        view_context = json.loads(payload["view"]["private_metadata"])
        switcher = self._view_submission_routes
        return switcher.get(callback_id, self.utils.NO_OP)(payload, view_context)

    def _view_closed_consumer(self, payload):
        view = payload["view"]
        callback_id = payload["view"]["callback_id"]
        switcher = self._view_closed_routes
        return switcher.get(callback_id, self.utils.NO_OP)(payload, view)

    def register_route(self, route_type, route_fn, name=None):
        """Register's interaction routes"""
        if route_type not in self._ROUTE_TYPES:
            raise ValueError(f"Invalid Route Type, only {', '.join(self._ROUTE_TYPES)} available")
        name = name if name else route_fn.__name__
        self.ROUTES.get(route_type).update({name: route_fn})
        return

    def register_block_set(self, block_set_fn, name=None):
        name = name if name else block_set_fn.__name__
        self.BLOCK_SETS.update({name: block_set_fn})
        return

    def slack_interaction_consumer(self, payload):
        """
        expects a payload to be sent
        example usage in django view
        payload = json.loads(request.data.get("payload"))
        process_output = slack_app.slack_interaction_consumer(payload)
        return Response(data=json.dumps(process_output))
        """
        switcher = {
            slack_consts.BLOCK_ACTIONS: self._block_action_consumer,
            slack_consts.BLOCK_SUGGESTION: self._block_suggestion_consumer,
            slack_consts.VIEW_SUBMISSION: self._view_submission_consumer,
            slack_consts.VIEW_CLOSED: self._view_closed_consumer,
        }

        return switcher.get(payload["type"], self.utils.NO_OP)(payload)

    def slack_event_request_consumer(self, request):
        if request.data.get("type") == slack_consts.EVENT_URL_VERIFICATION:
            event_type = "challenge"
            event_tab = None
            event_data = request.data.get("challenge")
        else:
            event_data = request.data.get("event", {})
            event_type = event_data.get("type")
            event_tab = event_data.get("tab")

        return dict(event_type=event_type, event_tab=event_tab, event_data=event_data)

    def get_block_set(self, set_name, context={}, *args, **kwargs):
        try:
            return self.BLOCK_SETS.get(set_name)(context, *args, **kwargs)
        except TypeError:
            raise TypeError("Block Set not found or not registered")

    def workspace_install_link(self, team_id=None, state=None):
        params = dict(
            redirect_uri=self.REDIRECT_URL,
            client_id=self.CLIENT_ID,
            scope=",".join(self.BOT_SCOPES),
        )
        if state:
            params["state"] = state
        if team_id:
            params["team_id"] = team_id
        return self.SLACK_OAUTH_AUTHORIZE_ROOT + "?" + urlencode(params)

    def user_install_link(self, team_id=None, state=None):
        params = dict(
            redirect_uri=self.REDIRECT_URL,
            client_id=self.CLIENT_ID,
            user_scope=",".join(self.USER_SCOPES),
        )
        if state:
            params["state"] = state
        if team_id:
            params["team_id"] = team_id
        return self.SLACK_OAUTH_AUTHORIZE_ROOT + "?" + urlencode(params)

    def generic_request(self, url, data, access_token=None):
        original_data = data
        res = requests.post(
            url,
            data=json.dumps(data),
            headers=self.auth_headers(access_token) if access_token else self.JSON_HEADERS,
        )
        return self._handle_response(res, blocks=original_data.get("blocks"))

    def open_user_conversation_channel(self, user_slack_id, access_token):
        """
        Request the Slack Channel ID for a 1:1 conversation
        between the user and bot
        """
        url = self.SLACK_API_ROOT + self.CONVERSATIONS_OPEN
        data = {"users": user_slack_id}
        res = requests.post(
            url,
            data=json.dumps(data),
            headers=self.auth_headers(access_token),
        )
        return self._handle_response(res, blocks=[])

    def send_channel_message(self, channel, access_token, text="Slack Message", block_set=[]):
        """
        Posts a message to a channel
        """
        url = self.SLACK_API_ROOT + self.POST_MESSAGE
        data = {}
        data["channel"] = channel
        data["text"] = text
        data["blocks"] = block_set

        res = requests.post(
            url,
            data=json.dumps(data),
            headers=self.auth_headers(access_token),
        )
        return self._handle_response(res, blocks=block_set)

    def publish_view(self, slack_id, access_token, view):
        """
        Publishes a view to the user's home tab
        slack_id: user slack id
        """
        url = self.SLACK_API_ROOT + self.VIEWS_PUBLISH
        data = {}
        data["user_id"] = slack_id
        data["view"] = view

        res = requests.post(
            url,
            data=json.dumps(data),
            headers=self.auth_headers(access_token),
        )
        return self._handle_response(res, data)

    def send_ephemeral_message(self, channel, access_token, slack_id, text="Slack Message", block_set=[]):
        """
        Posts a message to DM channel.
        *Channel and User are required for ephemeral messages
        """
        url = self.SLACK_API_ROOT + self.POST_EPHEMERAL
        data = {}
        data["channel"] = channel
        data["text"] = text
        data["blocks"] = block_set
        data["user"] = slack_id

        res = requests.post(
            url,
            data=json.dumps(data),
            headers=self.auth_headers(access_token),
        )
        return self._handle_response(res, blocks=block_set)

    def update_channel_message(self, channel, message_timestamp, access_token, text="Managr", block_set=[]):
        """
        Updates a message in DM.
        """
        url = self.SLACK_API_ROOT + self.CHAT_UPDATE
        data = {}
        data["channel"] = channel
        data["ts"] = message_timestamp

        data["text"] = text
        data["blocks"] = block_set
        res = requests.post(
            url,
            data=json.dumps(data),
            headers=self.auth_headers(access_token),
        )
        return self._handle_response(res, blocks=block_set if block_set else [])

    @property
    def utils(self):
        return SlackUtils()

    @property
    def block_builder(self):
        return BlockBuilder()

    @property
    def constants(self):
        return

    @property
    def _action_routes(self):
        return self.ROUTES.get(slack_consts.BLOCK_ACTIONS)

    @property
    def _suggestion_routes(self):
        return self.ROUTES.get(slack_consts.BLOCK_SUGGESTION)

    @property
    def _view_submission_routes(self):
        return self.ROUTES.get(slack_consts.VIEW_SUBMISSION)

    @property
    def _view_closed_routes(self):
        return self.ROUTES.get(slack_consts.VIEW_CLOSED)



class SlackUtils:
    def NO_OP(self, *args):
        """
        No operation.
        """
        pass

    def action_with_params(self, action, params=[]):
        """
        Max length of return is 255 characters.
        Slack will return a 200 but not display UI
        if action_id is greater than this limit.
        """
        if not isinstance(action, str):
            raise TypeError("action must be str")
        if not isinstance(params, list):
            raise TypeError("params must be list")
        output = action + "?" + "&".join(params)
        if len(output) > 255:
            raise ValueError("action_id would be longer than 255 characters")
        return output

    def process_action_id(self, action_string):
        output = {}
        x = action_string.split("?")
        output["true_id"] = x[0]
        output["params"] = {}
        if len(x) > 1:
            ps_list = x[1].split("&")
            for param_str in ps_list:
                b = param_str.split("=")
                output["params"][b[0]] = b[1]
        return output
